import * as http from 'http';
import * as https from 'https';
import { constants } from 'http2';
import { pluralize } from 'inflection';
import * as v from 'valibot';
import { SerializerPair } from './serializers';
import {
	handleCreateItem, handleDeleteItem, handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handleHasMethodAndUrl, handlePatchItem,
} from './handlers';
import Negotiator from 'negotiator';
import {getMethod, getUrl} from './utils';
import {EncodingPair} from './encodings';
import * as en from './languages/en';
import * as utf8 from './encodings/utf-8';
import * as applicationJson from './serializers/application/json';
import {Language} from './common';

// TODO separate frontend and backend factory methods

export interface DataSource<T = object, Q = object> {
	initialize(): Promise<unknown>;
	getTotalCount?(query?: Q): Promise<number>;
	getMultiple(query?: Q): Promise<T[]>;
	getById(id: string): Promise<T | null>;
	getSingle?(query?: Q): Promise<T | null>;
	create(data: T): Promise<T>;
	delete(id: string): Promise<unknown>;
	emplace(id: string, data: T): Promise<[T, boolean]>;
	patch(id: string, data: Partial<T>): Promise<T | null>;
}

export interface ApplicationParams {
	name: string;
	dataSource?: (resource: Resource) => DataSource;
}

interface ResourceState<IdAttr extends string = string, IdSchema extends v.BaseSchema = v.BaseSchema> {
	idAttr: IdAttr;
	itemName: string;
	collectionName: string;
	routeName: string;
	idConfig: ResourceIdConfig<IdSchema>;
	fullTextAttrs: Set<string>;
	canCreate: boolean;
	canFetchCollection: boolean;
	canFetchItem: boolean;
	canPatch: boolean;
	canEmplace: boolean;
	canDelete: boolean;
}

export interface Resource<
	ResourceSchema extends v.BaseSchema = v.BaseSchema,
	IdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
> {
	newId(dataSource: DataSource): string | number | unknown;
	schema: ResourceSchema;
	state: ResourceState<IdAttr, IdSchema>;
	id<NewIdAttr extends IdAttr, TheIdSchema extends v.BaseSchema>(
		newIdAttr: NewIdAttr,
		params: ResourceIdConfig<TheIdSchema>
	): Resource<ResourceSchema, NewIdAttr, TheIdSchema>;
	fullText(fullTextAttr: string): this;
	name(n: string): this;
	collection(n: string): this;
	route(n: string): this;
	canFetchCollection(b?: boolean): this;
	canFetchItem(b?: boolean): this;
	canCreate(b?: boolean): this;
	canPatch(b?: boolean): this;
	canEmplace(b?: boolean): this;
	canDelete(b?: boolean): this;
}

export interface ResourceWithDataSource<T extends v.BaseSchema> extends Resource<T> {
	dataSource: DataSource;
}

interface GenerationStrategy {
	(dataSource: DataSource, ...args: unknown[]): Promise<string | number | unknown>;
}

interface ResourceIdConfig<IdSchema extends v.BaseSchema> {
	generationStrategy: GenerationStrategy;
	serialize: (id: unknown) => string;
	deserialize: (id: string) => v.Output<IdSchema>;
	schema: IdSchema;
}

const getAllowedMiddlewares = <T extends v.BaseSchema>(resource: Resource<T>, mainResourceId: string) => {
	const middlewares = [] as [string, Middleware][];
	if (mainResourceId === '') {
		if (resource.state.canFetchCollection) {
			middlewares.push(['GET', handleGetCollection]);
		}
		if (resource.state.canCreate) {
			middlewares.push(['POST', handleCreateItem]);
		}
		return middlewares;
	}

	if (resource.state.canFetchItem) {
		middlewares.push(['GET', handleGetItem]);
	}
	if (resource.state.canEmplace) {
		middlewares.push(['PUT', handleEmplaceItem]);
	}
	if (resource.state.canPatch) {
		middlewares.push(['PATCH', handlePatchItem]);
	}
	if (resource.state.canDelete) {
		middlewares.push(['DELETE', handleDeleteItem]);
	}

	return middlewares;
};

export const resource = <
	ResourceSchema extends v.BaseSchema,
	IdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
>(schema: ResourceSchema): Resource<ResourceSchema, IdAttr, IdSchema> => {
	const resourceState = {
		fullTextAttrs: new Set<string>(),
		canCreate: false,
		canFetchCollection: false,
		canFetchItem: false,
		canPatch: false,
		canEmplace: false,
		canDelete: false,
	} as ResourceState<IdAttr, IdSchema>;

	return {
		get state(): ResourceState<IdAttr, IdSchema> {
			return Object.freeze({
				...resourceState
			}) as unknown as ResourceState<IdAttr, IdSchema>;
		},
		canFetchCollection(b = true) {
			resourceState.canFetchCollection = b;
			return this;
		},
		canFetchItem(b = true) {
			resourceState.canFetchItem = b;
			return this;
		},
		canCreate(b = true) {
			resourceState.canCreate = b;
			return this;
		},
		canPatch(b = true) {
			resourceState.canPatch = b;
			return this;
		},
		canEmplace(b = true) {
			resourceState.canEmplace = b;
			return this;
		},
		canDelete(b = true) {
			resourceState.canDelete = b;
			return this;
		},
		id<NewIdAttr extends IdAttr, NewIdSchema extends IdSchema>(newIdAttr: NewIdAttr, params: ResourceIdConfig<NewIdSchema>) {
			resourceState.idAttr = newIdAttr;
			resourceState.idConfig = params;
			return this as Resource<ResourceSchema, NewIdAttr, NewIdSchema>;
		},
		newId(dataSource: DataSource) {
			return resourceState?.idConfig?.generationStrategy?.(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (
				schema.type === 'object'
				&& (
					schema as unknown as v.ObjectSchema<
						Record<string, v.BaseSchema>,
						undefined,
						Record<string, string>
					>
				)
					.entries[fullTextAttr]?.type === 'string'
			) {
				resourceState.fullTextAttrs?.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name(n: string) {
			resourceState.itemName = n;
			resourceState.collectionName = resourceState.collectionName ?? pluralize(n).toLowerCase();
			resourceState.routeName = resourceState.routeName ?? resourceState.collectionName;
			return this;
		},
		collection(n: string) {
			resourceState.collectionName = n;
			resourceState.routeName = resourceState.routeName ?? n;
			return this;
		},
		route(n: string) {
			resourceState.routeName = n;
			return this;
		},
		get idAttr() {
			return resourceState.idAttr;
		},
		get collectionName() {
			return resourceState.collectionName;
		},
		get itemName() {
			return resourceState.itemName;
		},
		get routeName() {
			return resourceState.routeName;
		},
		get schema() {
			return schema;
		},
	} as Resource<ResourceSchema, IdAttr, IdSchema>;
};

type RequestListenerWithReturn<
	P extends unknown = unknown,
	Q extends typeof http.IncomingMessage = typeof http.IncomingMessage,
	R extends typeof http.ServerResponse = typeof http.ServerResponse
> = (
	...args: Parameters<http.RequestListener<Q, R>>
) => P;

interface HandlerState {
	handled: boolean;
}

export interface ApplicationState {
	resources: Set<ResourceWithDataSource<any>>;
	languages: Set<Language>;
	serializers: Set<SerializerPair>;
	encodings: Set<EncodingPair>;
}

export interface BackendState {
	fallback: {
		language: string;
		encoding: string;
		serializer: string;
	}
	errorHeaders: {
		language?: string;
		encoding?: string;
		serializer?: string;
	}
	showTotalItemCountOnGetCollection: boolean;
	throws404OnDeletingNotFound: boolean;
	checksSerializersOnDelete: boolean;
	showTotalItemCountOnCreateItem: boolean;
}

interface MiddlewareArgs<T extends v.BaseSchema> {
	handlerState: HandlerState;
	backendState: BackendState;
	appState: ApplicationState;
	appParams: ApplicationParams;
	serverParams: CreateServerParams;
	responseBodyLanguage: Language;
	responseBodyEncoding: EncodingPair;
	responseBodyMediaType: SerializerPair;
	errorResponseBodyLanguage: Language;
	errorResponseBodyEncoding: EncodingPair;
	errorResponseBodyMediaType: SerializerPair;
	resource: ResourceWithDataSource<T>;
	resourceId: string;
	query: URLSearchParams;
}

export interface Middleware {
	<T extends v.BaseSchema = v.BaseSchema>(args: MiddlewareArgs<T>): RequestListenerWithReturn<HandlerState | Promise<HandlerState>>
}

interface CreateServerParams {
	baseUrl?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
}

export interface Backend {
	showTotalItemCountOnGetCollection(b?: boolean): this;
	showTotalItemCountOnCreateItem(b?: boolean): this;
	checksSerializersOnDelete(b?: boolean): this;
	throws404OnDeletingNotFound(b?: boolean): this;
	createServer(serverParams?: CreateServerParams): http.Server | https.Server;
}

export interface Client {
	setLanguage(languageCode: string): this;
	setEncoding(encoding: string): this;
	setContentType(contentType: string): this;
}

export interface Application {
	contentType(serializerPair: SerializerPair): this;
	language(language: Language): this;
	encoding(encodingPair: EncodingPair): this;
	resource<T extends v.BaseSchema>(resRaw: Partial<Resource<T>>): this;
	createBackend(): Backend;
	createClient(): Client;
}

export const application = (appParams: ApplicationParams): Application => {
	const appState: ApplicationState = {
		resources: new Set<ResourceWithDataSource<any>>(),
		languages: new Set<Language>(),
		serializers: new Set<SerializerPair>(),
		encodings: new Set<EncodingPair>(),
	};

	appState.languages.add(en);
	appState.encodings.add(utf8);
	appState.serializers.add(applicationJson);

	return {
		contentType(serializerPair: SerializerPair) {
			appState.serializers.add(serializerPair);
			return this;
		},
		encoding(encodingPair: EncodingPair) {
			appState.encodings.add(encodingPair);
			return this;
		},
		language(language: Language) {
			appState.languages.add(language);
			return this;
		},
		resource<T extends v.BaseSchema>(resRaw: Partial<Resource<T>>) {
			const res = resRaw as Partial<ResourceWithDataSource<T>>;
			res.dataSource = res.dataSource ?? appParams.dataSource?.(res as Resource);
			if (typeof res.dataSource === 'undefined') {
				throw new Error(`Resource ${res.state!.itemName} must have a data source.`);
			}
			appState.resources.add(res as ResourceWithDataSource<T>);
			return this;
		},
		createClient(): Client {
			const clientState = {
				contentType: applicationJson.name,
				encoding: utf8.name,
				language: en.name as string
			};

			return {
				setContentType(contentType: string) {
					clientState.contentType = contentType;
					return this;
				},
				setEncoding(encoding: string) {
					clientState.encoding = encoding;
					return this;
				},
				setLanguage(languageCode: string) {
					clientState.language = languageCode;
					return this;
				}
			} satisfies Client;
		},
		createBackend(): Backend {
			const backendState: BackendState = {
				fallback: {
					language: en.name,
					encoding: utf8.name,
					serializer: applicationJson.name
				},
				errorHeaders: {
					// undefined follows user accept headers strictly
					//
					language: undefined,
					encoding: undefined,
					serializer: undefined,
				},
				showTotalItemCountOnGetCollection: false,
				showTotalItemCountOnCreateItem: false,
				throws404OnDeletingNotFound: false,
				checksSerializersOnDelete: false,
			};

			return {
				showTotalItemCountOnGetCollection(b = true) {
					backendState.showTotalItemCountOnGetCollection = b;
					return this;
				},
				showTotalItemCountOnCreateItem(b = true) {
					backendState.showTotalItemCountOnCreateItem = b;
					return this;
				},
				throws404OnDeletingNotFound(b = true) {
					backendState.throws404OnDeletingNotFound = b;
					return this;
				},
				checksSerializersOnDelete(b = true) {
					backendState.checksSerializersOnDelete = b;
					return this;
				},
				createServer(serverParams = {} as CreateServerParams) {
					const server = 'key' in serverParams && 'cert' in serverParams
						? https.createServer({
							key: serverParams.key,
							cert: serverParams.cert,
							requestTimeout: serverParams.requestTimeout
						})
						: http.createServer({
							requestTimeout: serverParams.requestTimeout
						});

					server.on('request', async (req, res) => {
						const method = getMethod(req);
						const baseUrl = serverParams.baseUrl ?? '';
						const { url, query } = getUrl(req, baseUrl);

						const negotiator = new Negotiator(req);
						const availableLanguages = Array.from(appState.languages);
						const languageCandidate = negotiator.language(availableLanguages.map((l) => l.name)) ?? backendState.fallback.language;
						const availableEncodings = Array.from(appState.encodings);
						const encodingCandidate = negotiator.encoding(availableEncodings.map((l) => l.name)) ?? backendState.fallback.encoding;
						const availableContentTypes = Array.from(appState.serializers);
						const contentTypeCandidate = negotiator.mediaType(availableContentTypes.map((l) => l.name)) ?? backendState.fallback.serializer;

						const fallbackMessageCollection = en as Language;
						const fallbackSerializerPair = applicationJson as SerializerPair;
						const fallbackEncoding = utf8 as EncodingPair;

						const errorLanguageCode = backendState.errorHeaders.language ?? backendState.fallback.language;
						const errorMessageCollection = availableLanguages.find((l) => l.name === errorLanguageCode) ?? fallbackMessageCollection;

						const errorContentType = backendState.errorHeaders.serializer ?? backendState.fallback.serializer;
						const errorSerializerPair = availableContentTypes.find((l) => l.name === errorContentType) ?? fallbackSerializerPair;

						const errorEncodingKey = backendState.errorHeaders.encoding ?? backendState.fallback.encoding;
						const errorEncoding = availableEncodings.find((l) => l.name === errorEncodingKey) ?? fallbackEncoding;

						// TODO refactor
						const currentLanguageMessages = availableLanguages.find((l) => l.name === languageCandidate);
						if (typeof currentLanguageMessages === 'undefined') {
							const data = errorMessageCollection.bodies.languageNotAcceptable();
							const responseRaw = errorSerializerPair.serialize(data);
							const response = errorEncoding.encode(responseRaw);
							res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
								'Content-Language': errorLanguageCode,
								'Content-Type': errorContentType,
								'Content-Encoding': errorEncodingKey,
							});
							res.statusMessage = errorMessageCollection.statusMessages.languageNotAcceptable();
							res.end(response);
							return;
						}

						const currentMediaType = availableContentTypes.find((l) => l.name === contentTypeCandidate);
						if (typeof currentMediaType === 'undefined') {
							const data = errorMessageCollection.bodies.languageNotAcceptable();
							const responseRaw = errorSerializerPair.serialize(data);
							const response = errorEncoding.encode(responseRaw);
							res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
								'Content-Language': errorLanguageCode,
								'Content-Type': errorContentType,
								'Content-Encoding': errorEncodingKey,
							});
							res.statusMessage = errorMessageCollection.statusMessages.mediaTypeNotAcceptable();
							res.end(response);
							return;
						}

						const responseBodyEncodingEntry = availableEncodings.find((l) => l.name === encodingCandidate);
						if (typeof responseBodyEncodingEntry === 'undefined') {
							const data = errorMessageCollection.bodies.languageNotAcceptable();
							const responseRaw = errorSerializerPair.serialize(data);
							const response = errorEncoding.encode(responseRaw);
							res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
								'Content-Language': errorLanguageCode,
								'Content-Type': errorContentType,
								'Content-Encoding': errorEncodingKey,
							});
							res.statusMessage = errorMessageCollection.statusMessages.encodingNotAcceptable();
							res.end(response);
							return;
						}

						const middlewareArgs: Omit<MiddlewareArgs<never>, 'resource' | 'resourceId'> = {
							handlerState: {
								handled: false
							},
							appState,
							appParams,
							backendState,
							serverParams,
							query,
							responseBodyEncoding: responseBodyEncodingEntry,
							responseBodyMediaType: currentMediaType,
							responseBodyLanguage: currentLanguageMessages,
							errorResponseBodyMediaType: errorSerializerPair,
							errorResponseBodyEncoding: errorEncoding,
							errorResponseBodyLanguage: errorMessageCollection,
						};

						const methodAndUrl = await handleHasMethodAndUrl(middlewareArgs as MiddlewareArgs<never>)(req, res);
						if (methodAndUrl.handled) {
							return;
						}

						if (url === '/') {
							const middlewareState = await handleGetRoot(middlewareArgs as MiddlewareArgs<never>)(req, res);
							if (middlewareState.handled) {
								return;
							}

							res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
								Allow: 'HEAD, GET'
							});
							res.statusMessage = errorMessageCollection.statusMessages.methodNotAllowed();
							res.end();
							return;
						}

						const [, resourceRouteName, resourceId = ''] = url.split('/');
						const resource = Array.from(appState.resources).find((r) => r.state!.routeName === resourceRouteName);
						if (typeof resource === 'undefined') {
							res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
							res.statusMessage = errorMessageCollection.statusMessages.urlNotFound();
							res.end();
							return;
						}

						const middlewares = getAllowedMiddlewares(resource, resourceId);
						const middlewareState = await middlewares
							.reduce(
								async (currentHandlerStatePromise, [middlewareMethod, middleware]) => {
									const currentHandlerState = await currentHandlerStatePromise;
									if (method !== middlewareMethod) {
										return currentHandlerState;
									}

									if (currentHandlerState.handled) {
										return currentHandlerState;
									}

									return middleware({
										...middlewareArgs,
										handlerState: currentHandlerState,
										resource,
										resourceId: resourceId,
									})(req, res);
								},
								Promise.resolve<HandlerState>({
									handled: false
								})
							);

						if (middlewareState.handled) {
							return;
						}

						if (middlewares.length > 0) {
							res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
								Allow: middlewares.map((m) => m[0]).join(', ')
							});
							res.statusMessage = errorMessageCollection.statusMessages.methodNotAllowed();
							res.end();
							return;
						}

						res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
						res.statusMessage = errorMessageCollection.statusMessages.urlNotFound();
						res.end();
						return;
					});

					return server;
				}
			} satisfies Backend;
		},
	};
};
