import { constants } from 'http2';
import * as v from 'valibot';
import {Middleware} from './core';
import {getBody, getDeserializerObjects} from './utils';
import {IncomingMessage, ServerResponse} from 'http';

export const handleHasMethodAndUrl: Middleware = ({
	errorResponseBodyLanguage,
}) => (req: IncomingMessage, res: ServerResponse) => {
	if (!req.method) {
		res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
			'Allow': 'HEAD, GET, POST, PUT, PATCH, DELETE',
			'Content-Language': errorResponseBodyLanguage.name,

		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.methodNotAllowed();
		res.end();
		return {
			handled: true
		};
	}

	if (!req.url) {
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.statusMessage = errorResponseBodyLanguage.statusMessages.badRequest();
		res.end();
		return {
			handled: true
		};
	}

	return {
		handled: false
	};
};

export const handleGetRoot: Middleware = ({
	appState,
	appParams,
	serverParams,
	responseBodyMediaType,
	responseBodyLanguage,
	responseBodyEncoding,
	errorResponseBodyLanguage,
}) => (_req: IncomingMessage, res: ServerResponse) => {
	const data = {
		name: appParams.name
	};

	let serialized;
	try {
		serialized = responseBodyMediaType.serialize(data);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let encoded;
	try {
		encoded = responseBodyEncoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	const theHeaders: Record<string, string> = {
		'Content-Type': responseBodyMediaType.name,
		'Content-Language': responseBodyLanguage.name,
		'Content-Encoding': responseBodyEncoding.name,
	};

	const registeredResources = Array.from(appState.resources);
	const availableResources = registeredResources.filter((r) => (
		r.canFetchCollection
		|| r.canCreate
	));

	if (availableResources.length > 0) {
		// we are using custom headers for links because the standard Link header
		// is referring to the document metadata (e.g. author, next page, etc)
		// https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link
		theHeaders['Link'] = availableResources
			.map((r) =>
				`<${serverParams.baseUrl}/${r.state.routeName}>; rel="related"; name="${r.state.collectionName}"`,
			)
			.join(', ');
	}
	res.writeHead(constants.HTTP_STATUS_OK, theHeaders);
	res.statusMessage = responseBodyLanguage.statusMessages.ok();
	res.end(encoded);
	return {
		handled: true
	};
};

export const handleGetCollection: Middleware = ({
	resource,
	responseBodyMediaType,
	responseBodyLanguage,
	responseBodyEncoding,
	errorResponseBodyLanguage,
	backendState,
	query,
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let data: v.Output<typeof resource.schema>[];
	let totalItemCount: number | undefined;
	try {
		// TODO querying mechanism
		data = await resource.dataSource.getMultiple(query); // TODO paginated responses per resource
		if (backendState.showTotalItemCountOnGetCollection && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount(query);
		}
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToFetchResourceCollection(resource);
		res.end();
		return {
			handled: true
		};
	}

	let serialized;
	try {
		serialized = responseBodyMediaType.serialize(data);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let encoded;
	try {
		encoded = responseBodyEncoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	const headers: Record<string, string> = {
		'Content-Type': responseBodyMediaType.name,
		'Content-Language': responseBodyLanguage.name,
		'Content-Encoding': responseBodyEncoding.name,
	};

	if (typeof totalItemCount !== 'undefined') {
		headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
	}

	res.writeHead(constants.HTTP_STATUS_OK, headers);
	res.statusMessage = responseBodyLanguage.statusMessages.resourceCollectionFetched(resource);
	res.end(encoded);
	return {
		handled: true
	};
};

export const handleGetItem: Middleware = ({
	resourceId,
	resource,
	responseBodyMediaType,
	responseBodyLanguage,
	responseBodyEncoding,
	errorResponseBodyLanguage,
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let data: v.Output<typeof resource.schema> | null = null;
	try {
		data = await resource.dataSource.getById(resourceId);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToFetchResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let serialized: string | null;
	try {
		serialized = data === null ? null : responseBodyMediaType.serialize(data);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let encoded;
	try {
		encoded = serialized === null ? null : responseBodyEncoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	if (encoded) {
		res.writeHead(constants.HTTP_STATUS_OK, {
			'Content-Type': responseBodyMediaType.name,
			'Content-Language': responseBodyLanguage.name,
			'Content-Encoding': responseBodyEncoding.name,
		});
		res.statusMessage = responseBodyLanguage.statusMessages.resourceFetched(resource)
		res.end(encoded);
		return {
			handled: true
		};
	}

	res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
		'Content-Language': errorResponseBodyLanguage.name,
	});
	res.statusMessage = errorResponseBodyLanguage.statusMessages.resourceNotFound(resource);
	res.end();
	return {
		handled: true
	};
};


export const handleDeleteItem: Middleware = ({
	resource,
	resourceId,
	responseBodyLanguage,
	errorResponseBodyLanguage,
	backendState,
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let existing: unknown | null;
	try {
		existing = await resource.dataSource.getById(resourceId);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToFetchResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	if (!existing && backendState.throws404OnDeletingNotFound) {
		res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.deleteNonExistingResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	try {
		if (existing) {
			await resource.dataSource.delete(resourceId);
		}
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToDeleteResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	res.writeHead(constants.HTTP_STATUS_NO_CONTENT, {
		'Content-Language': responseBodyLanguage.name,
	});
	res.statusMessage = responseBodyLanguage.statusMessages.resourceDeleted(resource);
	res.end();
	return {
		handled: true
	};
};

export const handlePatchItem: Middleware = ({
	appState,
	resource,
	resourceId,
	responseBodyMediaType,
	responseBodyLanguage,
	responseBodyEncoding,
	errorResponseBodyLanguage,
	errorResponseBodyMediaType,
	errorResponseBodyEncoding,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.writeHead(constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToDeserializeRequest();
		res.end();
		return {
			handled: true
		};
	}

	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let existing: unknown | null;
	try {
		existing = await resource.dataSource.getById(resourceId);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToFetchResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	if (!existing) {
		res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.patchNonExistingResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			schema.type === 'object'
				? v.partial(
					schema as v.ObjectSchema<any>,
					(schema as v.ObjectSchema<any>).rest,
					(schema as v.ObjectSchema<any>).pipe
				)
				: schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		const headers: Record<string, string> = {
			'Content-Language': responseBodyLanguage.name,
		};
		if (!Array.isArray(err.issues)) {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, headers)
			res.statusMessage = errorResponseBodyLanguage.statusMessages.invalidResourcePatch(resource);
			res.end();
			return {
				handled: true,
			};
		}
		// TODO better error reporting, localizable messages
		// TODO handle error handlers' errors
		const serialized = errorResponseBodyMediaType.serialize(
			err.issues.map((i) => (
				`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
			)),
		);
		const encoded = errorResponseBodyEncoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {
			...headers,
			'Content-Type': errorResponseBodyMediaType.name,
			'Content-Encoding': errorResponseBodyEncoding.name,
		})
		res.statusMessage = errorResponseBodyLanguage.statusMessages.invalidResourcePatch(resource);
		res.end(encoded);
		return {
			handled: true,
		};
	}

	const params = bodyDeserialized as Record<string, unknown>;
	let newObject: v.Output<typeof resource.schema> | null;
	try {
		newObject = await resource.dataSource.patch(resourceId, params);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToPatchResource(resource);
		res.end();
		return {
			handled: true,
		};
	}

	let serialized;
	try {
		serialized = responseBodyMediaType.serialize(newObject);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let encoded;
	try {
		encoded = responseBodyEncoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	res.writeHead(constants.HTTP_STATUS_OK, {
		'Content-Type': responseBodyMediaType.name,
		'Content-Language': responseBodyLanguage.name,
		'Content-Encoding': responseBodyEncoding.name,
	});
	res.statusMessage = responseBodyLanguage.statusMessages.resourcePatched(resource);
	res.end(encoded);
	return {
		handled: true
	};

	// TODO finish the rest of the handlers!!!
};

export const handleCreateItem: Middleware = ({
	appState,
	serverParams,
	backendState,
	responseBodyMediaType,
	responseBodyLanguage,
	responseBodyEncoding,
	errorResponseBodyLanguage,
	errorResponseBodyMediaType,
	errorResponseBodyEncoding,
	resource,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.writeHead(constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToDeserializeRequest();
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			resource.schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		const headers: Record<string, string> = {
			'Content-Language': errorResponseBodyLanguage.name,
		};
		if (!Array.isArray(err.issues)) {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, headers)
			res.statusMessage = errorResponseBodyLanguage.statusMessages.invalidResource(resource);
			res.end();
			return {
				handled: true,
			};
		}
		// TODO better error reporting, localizable messages
		// TODO handle error handlers' errors
		const serialized = errorResponseBodyMediaType.serialize(
			err.issues.map((i) => (
				`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
			)),
		);
		const encoded = errorResponseBodyEncoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {
			...headers,
			'Content-Type': errorResponseBodyMediaType.name,
			'Content-Encoding': errorResponseBodyEncoding.name,
		})
		res.statusMessage = errorResponseBodyLanguage.statusMessages.invalidResource(resource);
		res.end(encoded);
		return {
			handled: true,
		};
	}

	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	//v.Output<typeof resource.schema>

	let newId;
	let params: v.Output<typeof resource.schema>;
	try {
		newId = await resource.newId(resource.dataSource);
		params = bodyDeserialized as Record<string, unknown>;
		params[resource.state.idAttr] = newId;
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToGenerateIdFromResourceDataSource(resource);
		res.end();
		return {
			handled: true,
		};
		// noop
		// TODO
	}

	let newObject;
	let totalItemCount: number | undefined;
	try {
		newObject = await resource.dataSource.create(params);
		if (backendState.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount();
		}
	} catch {
		// noop
		// TODO
	}

	let serialized;
	try {
		serialized = responseBodyMediaType.serialize(newObject);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let encoded;
	try {
		encoded = responseBodyEncoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	const headers: Record<string, string> = {
		'Content-Type': responseBodyMediaType.name,
		'Content-Language': responseBodyLanguage.name,
		'Content-Encoding': responseBodyEncoding.name,
		'Location': `${serverParams.baseUrl}/${resource.state.routeName}/${newId}`
	};
	if (typeof totalItemCount !== 'undefined') {
		headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
	}
	res.writeHead(constants.HTTP_STATUS_CREATED, headers);
	res.statusMessage = responseBodyLanguage.statusMessages.resourceCreated(resource);
	res.end(encoded);
	return {
		handled: true
	};
}

export const handleEmplaceItem: Middleware = ({
	appState,
	serverParams,
	responseBodyMediaType,
	responseBodyLanguage,
	responseBodyEncoding,
	errorResponseBodyLanguage,
	errorResponseBodyMediaType,
	errorResponseBodyEncoding,
	resource,
	resourceId,
	backendState,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.writeHead(constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToDeserializeRequest();
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		const schema = resource.schema.type === 'object' ? resource.schema as unknown as v.ObjectSchema<any> : resource.schema
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			schema.type === 'object'
				? v.merge([
					schema as v.ObjectSchema<any>,
					v.object({
						[resource.state.idAttr]: v.transform(
							v.any(),
							input => resource.state.idConfig.serialize(input),
							v.literal(resourceId)
						)
					})
				])
				: schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		const headers: Record<string, string> = {
			'Content-Language': errorResponseBodyLanguage.name,
		};
		if (!Array.isArray(err.issues)) {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, headers)
			res.statusMessage = errorResponseBodyLanguage.statusMessages.invalidResource(resource);
			res.end();
			return {
				handled: true,
			};
		}
		// TODO better error reporting, localizable messages
		// TODO handle error handlers' errors
		const serialized = errorResponseBodyMediaType.serialize(
			err.issues.map((i) => (
				`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
			)),
		);
		const encoded = errorResponseBodyEncoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {
			...headers,
			'Content-Type': errorResponseBodyMediaType.name,
			'Content-Encoding': errorResponseBodyEncoding.name,
		})
		res.statusMessage = errorResponseBodyLanguage.statusMessages.invalidResource(resource);
		res.end(encoded);
		return {
			handled: true,
		};
	}

	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let newObject: v.Output<typeof resource.schema>;
	let isCreated: boolean;
	try {
		const params = bodyDeserialized as Record<string, unknown>;
		params[resource.state.idAttr] = resource.state.idConfig.deserialize(params[resource.state.idAttr] as string);
		[newObject, isCreated] = await resource.dataSource.emplace(resourceId, params);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEmplaceResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let serialized;
	try {
		serialized = responseBodyMediaType.serialize(newObject);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let encoded;
	try {
		encoded = responseBodyEncoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorResponseBodyLanguage.name,
		});
		res.statusMessage = errorResponseBodyLanguage.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	const headers: Record<string, string> = {
		'Content-Type': responseBodyMediaType.name,
		'Content-Language': responseBodyLanguage.name,
		'Content-Encoding': responseBodyEncoding.name,
	};
	let totalItemCount: number | undefined;
	if (backendState.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
		totalItemCount = await resource.dataSource.getTotalCount();
	}
	if (isCreated) {
		headers['Location'] = `${serverParams.baseUrl}/${resource.state.routeName}/${resourceId}`;
		if (typeof totalItemCount !== 'undefined') {
			headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
		}
	}
	res.writeHead(isCreated ? constants.HTTP_STATUS_CREATED : constants.HTTP_STATUS_OK, headers);
	res.statusMessage = (
		isCreated
			? responseBodyLanguage.statusMessages.resourceCreated(resource)
			: responseBodyLanguage.statusMessages.resourceReplaced(resource)
	);
	res.end(encoded);
	return {
		handled: true
	};
}
