import {IncomingMessage} from 'http';
import {SerializerPair} from './serializers';
import {BaseSchema, parseAsync} from 'valibot';
import { URL } from 'url';
import {EncodingPair} from './encodings';
import {ApplicationState} from './core';

export const getDeserializerObjects = (appState: ApplicationState, req: IncomingMessage) => {
	const availableSerializers = Array.from(appState.serializers);
	const availableEncodings = Array.from(appState.encodings);
	const deserializerPair = availableSerializers.find((l) => l.name === (req.headers['content-type'] ?? 'application/octet-stream'));
	const encodingPair = availableEncodings.find((l) => l.name === (req.headers['content-encoding'] ?? 'utf-8'));
	return {
		deserializerPair,
		encodingPair,
	};
};

export const getMethod = (req: IncomingMessage) => req.method!.trim().toUpperCase();

export const getUrl = (req: IncomingMessage, baseUrl?: string) => {
	const urlObject = new URL(req.url!, 'http://localhost');
	const urlWithoutBaseRaw = urlObject.pathname.slice(baseUrl?.length ?? 0);

	return {
		url: urlWithoutBaseRaw.length < 1 ? '/' : urlWithoutBaseRaw,
		query: urlObject.searchParams,
	};
}

export const getBody = (
	req: IncomingMessage,
	deserializer: SerializerPair,
	encodingPair: EncodingPair,
	schema: BaseSchema
) => new Promise((resolve, reject) => {
	let body = Buffer.from('');
	req.on('data', (chunk) => {
		body = Buffer.concat([body, chunk]);
	});
	req.on('end', async () => {
		const bodyStr = encodingPair.decode(body);
		try {
			const bodyDeserialized = await parseAsync(
				schema,
				deserializer.deserialize(bodyStr),
				{abortEarly: false},
			);
			resolve(bodyDeserialized);
		} catch (err) {
			reject(err);
		}
	});
});
